from guillotina import task_vars
from guillotina.utils import get_dotted_name
from guillotina_amqp.job import Job
from guillotina_amqp.state import TaskState
from guillotina_amqp.state import TaskStatus
from guillotina_amqp.tests.utils import _decorator_test_func
from guillotina_amqp.tests.utils import _test_asyncgen
from guillotina_amqp.tests.utils import _test_asyncgen_doubley
from guillotina_amqp.tests.utils import _test_asyncgen_invalid
from guillotina_amqp.tests.utils import _test_failing_func
from guillotina_amqp.tests.utils import _test_func
from guillotina_amqp.tests.utils import _test_long_func
from guillotina_amqp.utils import _run_object_task
from guillotina_amqp.utils import _yield_object_task
from guillotina_amqp.utils import add_task
from guillotina_amqp.utils import cancel_task

import asyncio
import json


async def test_add_task(
    dummy_request,
    rabbitmq_container,
    amqp_worker,
    amqp_channel,
    configured_state_manager,
):
    task_vars.request.set(dummy_request)
    ts = await add_task(_test_func, 1, 2)

    await ts.join(0.02)
    await asyncio.sleep(1)

    state = await ts.get_state()
    assert state["status"] == TaskStatus.FINISHED
    main_queue = await amqp_worker.queue_main(amqp_channel)
    assert main_queue["message_count"] == 0

    task_vars.request.set(None)


async def test_generator_tasks(
    dummy_request,
    rabbitmq_container,
    amqp_worker,
    amqp_channel,
    configured_state_manager,
):
    task_vars.request.set(dummy_request)
    ts = await add_task(_test_asyncgen, 1, 2)
    await ts.join(0.02)

    state = await ts.get_state()
    assert state["status"] == TaskStatus.FINISHED
    assert "Yellow" in state["eventlog"][-1]
    assert await ts.get_result() == [3]
    main_queue = await amqp_worker.queue_main(amqp_channel)
    assert main_queue["message_count"] == 0

    ts_invalid = await add_task(_test_asyncgen_invalid)
    await ts_invalid.join(0.02)
    assert await ts_invalid.get_result() is None

    ts_doubley = await add_task(_test_asyncgen_doubley, 128, 1024, 2048)
    await ts_doubley.join(0.02)
    assert await ts_doubley.get_result() == [128, 1024, 2048]

    task_vars.request.set(None)


async def test_run_task(dummy_request, rabbitmq_container, amqp_worker):
    task_vars.request.set(dummy_request)
    state = await add_task(_test_func, 1, 2)
    await state.join(0.01)
    await asyncio.sleep(0.1)  # prevent possible race condition here
    assert amqp_worker.total_run == 1
    assert await state.get_result() == 3
    task_vars.request.set(None)


async def test_task_from_service(rabbitmq_container, amqp_worker, container_requester):
    async with container_requester as requester:
        resp, _ = await requester("GET", "/db/guillotina/@foobar")
        state = TaskState(resp["task_id"])
        await state.join(0.01)
        assert await state.get_result() == 3
        await asyncio.sleep(0.1)  # prevent possible race condition here
        assert amqp_worker.total_run == 1


async def test_task_commits_data_from_service(
    rabbitmq_container, amqp_worker, container_requester
):
    async with container_requester as requester:
        await requester(
            "POST",
            "/db/guillotina",
            data=json.dumps({"@type": "Item", "id": "foobar", "title": "blah"}),
        )
        resp, _ = await requester("GET", "/db/guillotina/foobar/@foobar-write")
        state = TaskState(resp["task_id"])
        await state.join(0.01)
        assert await state.get_result() == "done!"
        await asyncio.sleep(0.1)  # prevent possible race condition here
        assert amqp_worker.total_run == 1
        resp, status = await requester("GET", "/db/guillotina/foobar")
        assert resp["title"] == "Foobar written"


async def test_async_gen_task_commits_data_from_service(
    configured_state_manager, rabbitmq_container, amqp_worker, container_requester
):
    async with container_requester as requester:
        await requester(
            "POST",
            "/db/guillotina",
            data=json.dumps({"@type": "Item", "id": "foobar", "title": "blah"}),
        )
        resp, _ = await requester(
            "GET", "/db/guillotina/foobar/@foobar-write-async-gen"
        )
        state = TaskState(resp["task_id"])
        await state.join(0.01)
        assert await state.get_status() == "finished"
        task_state = await state.get_state()
        assert len(task_state["eventlog"]) == 3
        assert await state.get_result() == ["one", "two", "final"]
        await asyncio.sleep(0.1)  # prevent possible race condition here
        assert amqp_worker.total_run == 1
        resp, status = await requester("GET", "/db/guillotina/foobar")
        assert resp["title"] == "CHANGED!"


async def test_cancels_long_running_task(
    dummy_request, rabbitmq_container, amqp_worker, configured_state_manager
):
    task_vars.request.set(dummy_request)
    # Add long running task
    ts = await _test_long_func(120)
    # Wait a bit and cancel
    await asyncio.sleep(1)
    success = await cancel_task(ts.task_id)
    assert success
    # Wait until worker sees task is cancelled and cancels the
    # asycio task
    await ts.join(0.2)

    # Check that the it was indeed cancelled
    state = await ts.get_state()
    assert state["status"] == TaskStatus.CANCELED
    await asyncio.sleep(0.1)  # prevent possible race condition here
    assert amqp_worker.total_run == 1
    task_vars.request.set(None)


async def test_decorator_task(dummy_request, rabbitmq_container, amqp_worker):
    task_vars.request.set(dummy_request)
    state = await _decorator_test_func(1, 2)
    data = await state.join(0.01)
    assert data["result"] == 3
    await asyncio.sleep(0.1)  # prevent possible race condition here
    assert amqp_worker.total_run == 1
    assert await state.get_status() == "finished"
    assert await state.get_result() == 3
    task_vars.request.set(None)


async def test_errored_job_should_be_published_to_delayed_queue(
    dummy_request, rabbitmq_container, amqp_worker, amqp_channel
):
    task_vars.request.set(dummy_request)
    ts = await _test_failing_func()
    # wait for it to finish
    await ts.join(0.1)
    assert amqp_worker.total_run == 1
    await amqp_worker.join()
    state = await ts.get_state()
    assert state["status"] == TaskStatus.ERRORED
    assert state["job_retries"] == 1

    task_id = state["job_data"]["task_id"]

    # Check that the job has been moved to the delay queue and
    # verify the task id

    delayed = await amqp_worker.queue_delayed(amqp_channel)
    assert delayed["message_count"] == 1

    async def callback(channel, body, envelope, properties):
        decoded = json.loads(body)
        assert decoded["task_id"] == task_id

    await amqp_channel.basic_consume(callback, queue_name=delayed["queue"])
    task_vars.request.set(None)


async def test_worker_retries_should_not_exceed_the_limit(
    dummy_request, rabbitmq_container, amqp_worker, amqp_channel
):
    task_vars.request.set(dummy_request)

    # Add failing function and wait for it to finish
    ts = await _test_failing_func()

    # Fake job retries to max
    max_retries = amqp_worker.max_task_retries
    await amqp_worker.state_manager.update(ts.task_id, {"job_retries": max_retries})
    # Wait for it to finish
    await ts.join(0.1)
    # Check that worker ran only one
    assert amqp_worker.total_run == 1
    await amqp_worker.join()

    # Check that it retries up to max
    state = await ts.get_state()
    retried = state["job_retries"]
    assert retried == max_retries

    # Check that it went to error queue
    main_queue = await amqp_worker.queue_main(amqp_channel)
    errored_queue = await amqp_worker.queue_errored(amqp_channel)
    assert main_queue["consumer_count"] == 1
    assert errored_queue["message_count"] == 1

    task_vars.request.set(None)


def test_job_function_name():
    data = {"func": get_dotted_name(_run_object_task), "args": ["my.func.foobar"]}
    job = Job(None, data, None, None)
    assert job.function_name == "run_object_task/my.func.foobar"

    data = {"func": get_dotted_name(_yield_object_task), "args": ["my.func.foobar"]}
    job = Job(None, data, None, None)
    assert job.function_name == "yield_object_task/my.func.foobar"

    data = {"func": get_dotted_name(_test_func), "args": []}
    job = Job(None, data, None, None)
    assert job.function_name == "guillotina_amqp.tests.utils._test_func"



from aiohttp import web
from guillotina import glogging
from guillotina import task_vars
from guillotina.commands.server import ServerCommand
from guillotina.tests.utils import get_mocked_request
from guillotina_amqp.worker import Worker

import asyncio
import os
import threading


try:
    import prometheus_client
except ImportError:
    prometheus_client = None


logger = glogging.getLogger("guillotina_amqp")


async def prometheus_view(request):
    if prometheus_client is None:
        return None
    output = prometheus_client.exposition.generate_latest()
    return web.Response(text=output.decode("utf8"))


class EventLoopWatchdog(threading.Thread):
    """Takes care of exiting worker after specified loop no-activity
    timeout for the worker loop.

    This prevents a task from taking over the asynio loop forever and
    preventing other tasks to run.

    If a task hangs, the watchdog will exit the worker and unfinished jobs
    will be taken by other workers.
    """

    def __init__(self, loop, timeout):
        super().__init__()
        self.loop = loop
        self.timeout = timeout * 60  # In seconds
        # Start time
        self._time = loop.time()

    def check(self):
        # Elapsed time since last update
        diff = self.loop.time() - self._time

        if diff > self.timeout:
            logger.error(f"Exiting worker because no activity in {diff} seconds")
            os._exit(0)
        else:
            # Schedule a check again
            threading.Timer(self.timeout / 4, self.check).start()
            logger.debug(f"Last refreshed watchdog was {diff}s. ago")

    async def probe(self):
        """This method is used to trigger a context switching in the event
        loop and measure elapsed time between runs (10s)
        """
        while True:
            await asyncio.sleep(10)
            # Update the watchdog time
            self._time = self.loop.time()

    def run(self):
        self.loop.create_task(self.probe())
        threading.Timer(self.timeout / 4, self.check).start()


class WorkerCommand(ServerCommand):
    """Guillotina command to start a worker"""

    description = "AMQP worker"

    def get_parser(self):
        parser = super().get_parser()
        parser.add_argument(
            "--auto-kill-timeout",
            help="How long of no activity before we automatically kill process (in minutes)",
            type=int,
            default=-1,
        )
        parser.add_argument(
            "--max-running-tasks",
            help="Max simultaneous running tasks",
            type=int,
            default=None,
        )
        parser.add_argument(
            "--metrics-server",
            help="Launch an API server to expose metrics",
            default=False,
            action="store_true",
        )
        return parser

    def run(self, arguments, settings, app):
        loop = self.get_loop()
        if arguments.metrics_server:
            asyncio.ensure_future(
                self.run_worker(arguments, settings, app, loop=loop), loop=loop
            )
            port = arguments.port or settings.get("address", settings.get("port"))
            app = web.Application()
            app.router.add_get("/metrics", prometheus_view)
            web.run_app(app, port=port or 8080)
        else:
            loop.run_until_complete(self.run_worker(arguments, settings, app))

    async def run_worker(self, arguments, settings, app, loop=None):
        try:
            await self._run_worker(arguments, settings, app, loop)
        except Exception:
            logger.error("Error running worker. Exiting", exc_info=True)
            os._exit(0)

    async def _run_worker(self, arguments, settings, app, loop=None):
        self.request = get_mocked_request()
        task_vars.request.set(self.request)

        loop = loop or self.get_loop()

        # Run the actual worker in the same loop
        worker = Worker(self.request, loop, arguments.max_running_tasks)
        await worker.start()

        timeout = arguments.auto_kill_timeout
        if timeout > 0:
            # We need to run this outside the main loop and the current thread
            thread = EventLoopWatchdog(loop, timeout)
            thread.start()

        while True:
            # make this run forever...
            await asyncio.sleep(999999)



class TaskNotFinishedException(Exception):
    pass


class TaskNotFoundException(Exception):
    pass


class TaskAlreadyAcquired(Exception):
    pass


class TaskAlreadyCanceled(Exception):
    pass


class TaskAccessUnauthorized(Exception):
    pass


class TaskMaxRetriesReached(Exception):
    pass


class ObjectNotFoundException(Exception):
    pass



from guillotina import configure


# Add new permission
configure.permission("guillotina.ManageAMQP", "Manage guillotina amqp endpoints")

# Grant it to guillotina.Manager
configure.grant(permission="guillotina.ManageAMQP", role="guillotina.Manager")



from guillotina import app_settings
from guillotina import glogging
from guillotina_amqp import amqp
from guillotina_amqp.exceptions import TaskAlreadyAcquired
from guillotina_amqp.exceptions import TaskAlreadyCanceled
from guillotina_amqp.job import Job
from guillotina_amqp.state import get_state_manager
from guillotina_amqp.state import TaskState
from guillotina_amqp.state import TaskStatus
from guillotina_amqp.state import update_task_canceled
from guillotina_amqp.state import update_task_errored
from guillotina_amqp.state import update_task_finished
from guillotina_amqp.state import update_task_scheduled
from guillotina_amqp.utils import metric_measure

import asyncio
import json
import time
import os

try:
    from prometheus_client import Gauge
    from prometheus_client import Histogram

    amqp_running_jobs = Gauge(
        "amqp_running_jobs", "Number of AQMP running jobs in worker", []
    )

    amqp_job_duration = Histogram(
        "amqp_job_duration",
        "AMQP job duration histogram",
        ["dotted_name", "final_status", "container_id"],
    )

except ImportError:
    # Do not record metrics if prometheus_client not installed
    amqp_running_jobs = None
    amqp_job_duration = None


logger = glogging.getLogger("guillotina_amqp.worker")
default_delayed = 1000 * 60 * 2  # 2 minutes
default_errored = 1000 * 60 * 60 * 24 * 7 * 1  # 1 week


class Worker:
    """Workers hold an asyncio loop in which will run several tasks. It
    reads from RabbitMQ for new job descriptions and will run them in
    asyncio tasks.

    The worker is aware of a state manager in which he posts job
    results.

    """

    sleep_interval = 0.1
    last_activity = time.time()
    update_status_interval = 20
    total_run = 0
    total_errored = 0
    max_task_retries = 5
    _status_task = None

    def __init__(self, request=None, loop=None, max_size=None, check_activity=True):
        self.request = request
        self.loop = loop
        self._running = []
        self._max_running = int(
            max_size or app_settings["amqp"].get("max_running_tasks", 5)
        )
        self._closing = False
        self._state_manager = None
        self._state_ttl = int(app_settings["amqp"]["state_ttl"])
        self._check_activity = check_activity

        # RabbitMQ queue names defined here
        self.MAIN_EXCHANGE = app_settings["amqp"]["exchange"]
        self.QUEUE_MAIN = app_settings["amqp"]["queue"]
        self.QUEUE_ERRORED = app_settings["amqp"]["queue"] + "-error"
        self.QUEUE_DELAYED = app_settings["amqp"]["queue"] + "-delay"
        self.TTL_ERRORED = app_settings["amqp"].get("errored_ttl_ms", default_errored)
        self.TTL_DELAYED = app_settings["amqp"].get("delayed_ttl_ms", default_delayed)

    @property
    def state_manager(self):
        if self._state_manager is None:
            self._state_manager = get_state_manager()
        return self._state_manager

    @property
    def num_running(self):
        """Returns the number of currently running jobs"""
        return len(self._running)

    async def handle_queued_job(self, channel, body, envelope, properties):
        """Callback triggered when there is a new job in the job channel (e.g:
        a new task in a rabbitmq queue)

        """
        logger.debug(f"Queued job {body}")

        # Deserialize job description
        if not isinstance(body, str):
            body = body.decode("utf-8")
        data = json.loads(body)

        # Set status to scheduled
        task_id = data["task_id"]
        dotted_name = data["func"]

        await update_task_scheduled(self.state_manager, task_id, eventlog=[])
        logger.info(f"Received task: {task_id}: {dotted_name}")

        self.measure_running_jobs(len(self._running))

        # Block if we reached maximum number of running tasks
        while len(self._running) >= self._max_running:
            logger.info(f"Max running tasks reached: {self._max_running}")
            await asyncio.sleep(self.sleep_interval)
            self.last_activity = time.time()

        # Create job object
        self.last_activity = time.time()
        job = Job(self.request, data, channel, envelope)
        # Get the redis lock on the task so no other worker takes it
        _id = job.data["task_id"]
        ts = TaskState(_id)

        # Cancelation
        if await ts.is_canceled():
            logger.warning(f"Task {_id} has already been canceled")
            raise TaskAlreadyCanceled(_id)

        try:
            await ts.acquire()
        except TaskAlreadyAcquired:
            logger.warning(f"Task {_id} is already running in another worker")
            # TODO: ACK task

        # Record job's data into global state
        await self.state_manager.update(task_id, {"job_data": job.data})

        # Add the task to the loop and start it
        task = self.loop.create_task(job())
        task._job = job
        job.task = task
        self._running.append(task)
        task.add_done_callback(self._task_done_callback)

    async def _handle_canceled(self, task):
        task_id = task._job.data["task_id"]
        # ACK to main queue to it is not scheduled anymore
        await task._job.channel.basic_client_ack(
            delivery_tag=task._job.envelope.delivery_tag
        )

        # Set status to cancelled
        await update_task_canceled(
            self.state_manager, task_id, task=task, ttl=self._state_ttl
        )

        self.measure_job_duration(task._job, TaskStatus.CANCELED)

    async def _handle_max_retries_reached(self, task):
        task_id = task._job.data["task_id"]
        logger.warning(f"Task {task_id} reached max {self.max_task_retries} retries")

        # Send NACK, as we exceeded the number of retrials
        await task._job.channel.basic_client_nack(
            delivery_tag=task._job.envelope.delivery_tag, multiple=False, requeue=False
        )

        # Update status to errored with the traceback
        await update_task_errored(
            self.state_manager, task_id, task=task, ttl=self._state_ttl
        )

        self.measure_job_duration(task._job, TaskStatus.ERRORED)

    async def _handle_retry(self, task, current_retries):
        task_id = task._job.data["task_id"]
        channel = task._job.channel
        logger.debug(f"handle_retry: task {task_id} retries {current_retries}")
        # Increment retry count
        await update_task_errored(
            self.state_manager,
            task_id,
            task=task,
            ttl=self._state_ttl,
            job_retries=current_retries + 1,
        )

        # Publish task data to delay queue
        await channel.publish(
            json.dumps(task._job.data),
            exchange_name=self.MAIN_EXCHANGE,
            routing_key=self.QUEUE_DELAYED,
            properties={"delivery_mode": 2},
        )
        # ACK to main queue so it doesn't timeout
        await channel.basic_client_ack(delivery_tag=task._job.envelope.delivery_tag)
        logger.info(f"Task {task_id} will be retried")

        self.measure_job_duration(task._job, TaskStatus.ERRORED)

    async def _handle_successful(self, task):
        task_id = task._job.data["task_id"]
        dotted_name = task._job.data["func"]

        # ACK rabbitmq: will make task disappear from rabbitmq
        await task._job.channel.basic_client_ack(
            delivery_tag=task._job.envelope.delivery_tag
        )

        # Update status with result
        await update_task_finished(
            self.state_manager,
            task_id,
            task=task,
            ttl=self._state_ttl,
            result=task.result(),
        )
        logger.info(f"Finished task: {task_id}: {dotted_name}")

        self.measure_job_duration(task._job, TaskStatus.FINISHED)

    @staticmethod
    def measure_job_duration(job, final_status):
        labels = {
            "dotted_name": job.function_name,
            "final_status": final_status,
            "container_id": job.data.get("container_id"),
        }
        duration = time.time() - job._started
        metric_measure(amqp_job_duration, duration, labels)

    @staticmethod
    def measure_running_jobs(currently_running):
        metric_measure(amqp_running_jobs, currently_running)

    def _task_done_callback(self, task):
        # We can't pass coroutines to add_done_callback so we have to
        # place it inside an ensure_future
        asyncio.ensure_future(self._task_callback(task))

    async def _task_callback(self, task):
        """This is called when a job finishes execution"""
        task_id = task._job.data["task_id"]
        self.total_run += 1

        try:
            result = task.result()
            logger.debug(f"Task data: {task._job.data}, result: {result}")
        except asyncio.CancelledError:
            logger.warning(f"Task got cancelled: {task._job.data}", exc_info=True)
            return await self._handle_canceled(task)
        except Exception:
            logger.error(f"Unhandled task exception: {task_id}", exc_info=True)
            # Error during execution of the task
            #
            # If max retries reached
            existing_data = await self.state_manager.get(task_id)
            retrials = existing_data.get("job_retries", 0)
            if retrials >= self.max_task_retries:
                return await self._handle_max_retries_reached(task)

            # Otherwise let task be retried
            return await self._handle_retry(task, retrials)
        else:
            # If task ran successfully, ACK main queue and finish
            return await self._handle_successful(task)
        finally:
            if task in self._running:
                self._running.remove(task)
            await self.state_manager.release(task_id)

    async def stop(self):
        self.cancel()
        await amqp.remove_connection()

    async def start(self):
        """Called on worker startup. Connects to the rabbitmq. Declares and
        configures the different queues.

        """
        channel, transport, protocol = await amqp.get_connection()

        # Declare main exchange
        await channel.exchange_declare(
            exchange_name=self.MAIN_EXCHANGE, type_name="direct", durable=True
        )

        # Declare errored queue and bind it
        await self.queue_errored(channel, passive=False)

        # Declare main queue and bind it
        await self.queue_main(channel, passive=False)

        # Declare delayed queue and bind it
        await self.queue_delayed(channel, passive=False)

        await channel.basic_qos(prefetch_count=self._max_running)

        # Configure task consume callback
        await channel.basic_consume(self.handle_queued_job, queue_name=self.QUEUE_MAIN)

        # Start task that will update status periodically
        self._status_task = asyncio.ensure_future(self.update_status())

        logger.warning(f"Subscribed to queue: {self.QUEUE_MAIN}")

    async def queue_main(self, channel, passive=True):
        """Declares the main queue for task messages. NACKed messages are sent
        to the errored queue.

        If passie is False, will additionally bind the queue to the
        exchange
        """
        resp = await channel.queue_declare(
            queue_name=self.QUEUE_MAIN,
            durable=True,
            passive=passive,
            arguments={
                "x-dead-letter-exchange": self.MAIN_EXCHANGE,
                "x-dead-letter-routing-key": self.QUEUE_ERRORED,
            },
        )
        if not passive:
            await channel.queue_bind(
                exchange_name=self.MAIN_EXCHANGE,
                queue_name=self.QUEUE_MAIN,
                routing_key=self.QUEUE_MAIN,
            )
        return resp

    async def queue_delayed(self, channel, passive=True):
        """Declares the queue for delayed tasks, which is used for failed
        tasks retrials. After self.TTL_DELAYED, tasks will be requeued
        to the main task queue.
        """
        resp = await channel.queue_declare(
            queue_name=self.QUEUE_DELAYED,
            durable=True,
            passive=passive,
            arguments={
                "x-dead-letter-exchange": self.MAIN_EXCHANGE,
                "x-dead-letter-routing-key": self.QUEUE_MAIN,
                "x-message-ttl": self.TTL_DELAYED,
            },
        )
        if not passive:
            await channel.queue_bind(
                exchange_name=self.MAIN_EXCHANGE,
                queue_name=self.QUEUE_DELAYED,
                routing_key=self.QUEUE_DELAYED,
            )
        return resp

    async def queue_errored(self, channel, passive=True):
        """Declares queue for errored tasks. Errored tasks will remain a
        limited period of time and then they will be lost.
        """
        resp = await channel.queue_declare(
            queue_name=self.QUEUE_ERRORED,
            durable=True,
            passive=passive,
            arguments={"x-message-ttl": self.TTL_ERRORED},
        )
        if not passive:
            await channel.queue_bind(
                exchange_name=self.MAIN_EXCHANGE,
                queue_name=self.QUEUE_ERRORED,
                routing_key=self.QUEUE_ERRORED,
            )
        return resp

    def cancel(self):
        """
        Cancels the worker (i.e: all its running tasks)
        """
        for task in self._running[:]:
            if not task.done():
                task.cancel()
            self._running.remove(task)
        if self._status_task is not None and not self._status_task.done():
            self._status_task.cancel()

    async def join(self):
        """
        Waits for all tasks to finish
        """
        while len(self._running) > 0:
            await asyncio.sleep(0.01)

    async def update_status(self):
        """Updates status for running tasks and kills running tasks that have
        been canceled.

        """
        while True:
            await asyncio.sleep(self.update_status_interval)

            if self._check_activity and len(self._running) == 0:
                diff = time.time() - self.last_activity
                if diff > 60 * 5:
                    logger.error(f"Exiting worker because no connection activity in {diff} seconds")
                    os._exit(0)

            for task in self._running:
                _id = task._job.data["task_id"]
                ts = TaskState(_id)

                # Still working on the job: refresh task lock
                await ts.refresh_lock()

            # Cancel local tasks that have been cancelled in global
            # state manager
            async for val in self.state_manager.cancelation_list():
                for task in self._running:
                    _id = task._job.data["task_id"]
                    if _id == val:
                        logger.warning(f"Canceling task {_id}")
                        if not task.done():
                            task.cancel()
                        self._running.remove(task)
                        await self._state_manager.clean_canceled(_id)



from guillotina import configure
from guillotina_amqp.interfaces import MessageType
from guillotina_amqp.utils import add_object_task
from guillotina_amqp.utils import add_task


async def task_foobar_yo(one, two, three="blah"):
    return one + two


async def task_object_write(ob, value):
    ob.title = value
    ob.register()
    return "done!"


async def task_object_write_generator(ob, values=None):
    for value in values or []:
        # This will be accumulated in the 'eventlog' key
        yield (MessageType.DEBUG, f"debug-{value}")
        # This will be accumulated in the 'result' key
        yield (MessageType.RESULT, value)
    if values:
        # Change the title of the object
        ob.title = "CHANGED!"
        ob.register()


@configure.service(name="@foobar", method="GET")
async def foobar(context, request):
    # Endpoint to be used in tests to add a function task
    return {"task_id": (await add_task(task_foobar_yo, 1, 2, three="hello!")).task_id}


@configure.service(name="@foobar-write", method="GET")
async def foobar_write(context, request):
    # Endpoint to be used in tests to add an object function task
    return {
        "task_id": (
            await add_object_task(task_object_write, context, "Foobar written")
        ).task_id
    }


@configure.service(name="@foobar-write-async-gen", method="GET")
async def foobar_write_async_gen(context, request):
    # Endpoint to be used in tests to add an object function task
    values = ["one", "two", "final"]
    return {
        "task_id": (
            await add_object_task(task_object_write_generator, context, values)
        ).task_id
    }



from guillotina import app_settings
from guillotina import glogging
from guillotina import task_vars
from guillotina.interfaces import Allow
from guillotina.interfaces import IAbsoluteURL
from guillotina.utils import get_authenticated_user
from guillotina.utils import get_content_path
from guillotina.utils import get_current_request
from guillotina.utils import get_dotted_name
from guillotina.utils import navigate_to
from guillotina.utils import resolve_dotted_name
from guillotina.utils.misc import get_current_container
from guillotina_amqp import amqp
from guillotina_amqp.exceptions import ObjectNotFoundException
from guillotina_amqp.interfaces import ITaskDefinition
from guillotina_amqp.state import get_state_manager
from guillotina_amqp.state import TaskState
from guillotina_amqp.state import update_task_scheduled

import aioamqp
import asyncio
import inspect
import json
import time
import uuid


logger = glogging.getLogger("guillotina_amqp.utils")


async def cancel_task(task_id):
    """It cancels a task by id. Returns wether it could be cancelled.

    """
    task = TaskState(task_id)
    success = await task.cancel()
    return success


def get_task_id_prefix():
    db = task_vars.db.get()
    container = task_vars.container.get()
    return "task:{}-{}-".format(db.id, container.id)


def generate_task_id():
    container = task_vars.container.get()
    if container is not None:
        return "{}{}".format(get_task_id_prefix(), str(uuid.uuid4()))
    return str(uuid.uuid4())


async def add_task(func, *args, _request=None, _retries=3, _task_id=None, **kwargs):
    """Given a function and its arguments, it adds it as a task to be ran
    by workers.
    """
    # Get the request and prepare request data
    if _request is None:
        _request = get_current_request()

    req_data = {
        "url": str(_request.url),
        "headers": dict(_request.headers),
        "method": _request.method,
        "annotations": getattr(_request, "annotations", {}),
    }
    user = get_authenticated_user()
    if user is not None:
        try:
            req_data["user"] = {
                "id": user.id,
                "roles": [
                    name for name, setting in user.roles.items() if setting == Allow
                ],
                "groups": user.groups,
                "headers": dict(_request.headers),
                "data": getattr(user, "data", {}),
            }
        except AttributeError:
            pass

    container = task_vars.container.get()
    if container is not None:
        req_data["container_url"] = IAbsoluteURL(container, _request)()

    if _task_id is None:
        task_id = generate_task_id()
    else:
        task_id = _task_id

    retries = 0
    while True:
        # Get the rabbitmq connection
        channel, transport, protocol = await amqp.get_connection()
        try:
            state = TaskState(task_id)
            dotted_name = get_dotted_name(func)
            db = task_vars.db.get()
            logger.info(f"Scheduling task: {task_id}: {dotted_name}")
            data = json.dumps(
                {
                    "func": dotted_name,
                    "args": args,
                    "kwargs": kwargs,
                    "db_id": getattr(db, "id", None),
                    "container_id": getattr(container, "id", None),
                    "req_data": req_data,
                    "task_id": task_id,
                }
            )
            # Publish task data on rabbitmq
            await channel.publish(
                data,
                exchange_name=app_settings["amqp"]["exchange"],
                routing_key=app_settings["amqp"]["queue"],
                properties={"delivery_mode": 2},
            )
            # Update tasks's global state
            state_manager = get_state_manager()
            await update_task_scheduled(state_manager, task_id, updated=time.time())
            logger.info(f"Scheduled task: {task_id}: {dotted_name}")
            return state
        except (aioamqp.AmqpClosedConnection, aioamqp.exceptions.ChannelClosed):
            await amqp.remove_connection()
            if retries >= _retries:
                raise
            retries += 1


async def _prepare_func(dotted_func, path, *args, **kwargs):
    container = get_current_container()
    try:
        ob = await navigate_to(container, path)
    except KeyError:
        logger.warning(f"Object in {path} not found")
        raise ObjectNotFoundException
    func = resolve_dotted_name(dotted_func)
    if ITaskDefinition.providedBy(func):
        func = func.func
    return ob, func


async def _run_object_task(dotted_func, path, *args, **kwargs):
    ob, func = await _prepare_func(dotted_func, path, *args, **kwargs)
    return await func(ob, *args, **kwargs)


async def _yield_object_task(dotted_func, path, *args, **kwargs):
    ob, func = await _prepare_func(dotted_func, path, *args, **kwargs)
    async for res in func(ob, *args, **kwargs):
        yield res


async def add_object_task(
    callable=None, ob=None, *args, _request=None, _retries=3, **kwargs
):
    superfunc = _run_object_task
    if inspect.isasyncgenfunction(callable):
        # async generators need to be yielded from
        superfunc = _yield_object_task
    return await add_task(
        superfunc,
        get_dotted_name(callable),
        get_content_path(ob),
        *args,
        _request=_request,
        _retries=_retries,
        **kwargs,
    )


class TimeoutLock(object):
    """Implements a Lock that can be acquired for """

    def __init__(self, worker_id):
        self._lock = asyncio.Lock()
        self.worker_id = worker_id

    async def acquire(self, ttl=-1):
        """If ttl is -1, lock will be acquired forever (or until someone
        manually releases it).

        Otherwise, it schedules an automatic release after the
        specified ttl.
        """
        if self.locked():
            # Already acquired
            raise False

        acquired = await self._lock.acquire()
        if not acquired:
            return False

        if ttl >= 0:
            asyncio.ensure_future(self._release_after(ttl))
        return True

    async def _release_after(self, some_time):
        await asyncio.sleep(some_time)
        self.release()

    def release(self):
        if self.locked():
            self._lock.release()

    def locked(self):
        return self._lock.locked()

    async def refresh_lock(self, ttl):
        # Overwrite old lock and acquire with new timeout
        self._lock = asyncio.Lock()
        return await self.acquire(ttl)


def metric_measure(metric, value, labels=None):
    if not metric:
        # Don't measure if prometheus client is not installed
        return

    try:
        labels = labels or {}
        try:
            labeled = metric.labels(**labels)
            try:
                labeled.observe(value)
            except AttributeError:
                labeled.set(value)
        except AttributeError:
            metric.set(value)
    except Exception:
        logger.error(f"Failed to measure metric", exc_info=True)
        pass



from functools import partial
from guillotina.transactions import get_transaction
from guillotina.utils import get_current_request
from guillotina_amqp.interfaces import ITaskDefinition
from guillotina_amqp.utils import add_object_task
from guillotina_amqp.utils import add_task
from zope.interface import implementer

import uuid


@implementer(ITaskDefinition)
class TaskDefinition:
    def __init__(self, func, retries=3):
        self.func = func
        self.retries = retries

    async def __call__(self, *args, _request=None, **kwargs):
        return await add_task(
            self.func, _request=_request, _retries=self.retries, *args, **kwargs
        )

    schedule = __call__

    def _get_request(self, request, kwargs):
        if request is None:
            if "request" in kwargs:
                request = kwargs["request"]
            else:
                request = get_current_request()
        return request

    def after_request(self, *args, _request=None, _name=None, **kwargs):
        request = self._get_request(_request, kwargs)
        kwargs["_request"] = request
        if _name is None:
            _name = str(uuid.uuid4())
        request.add_future(_name, partial(self.schedule, *args, **kwargs))

    def after_commit(self, *args, _request=None, **kwargs):
        request = self._get_request(_request, kwargs)
        txn = get_transaction(request)
        kwargs["_request"] = request
        txn.add_after_commit_hook(self.schedule, args=args, kwargs=kwargs)


class ObjectTaskDefinition(TaskDefinition):
    async def __call__(self, *args, _request=None, **kwargs):
        return await add_object_task(
            self.func, _request=_request, _retries=self.retries, *args, **kwargs
        )

    schedule = __call__


def task(func, retries=3):
    return TaskDefinition(func, retries)


def object_task(func, retries=3):
    return ObjectTaskDefinition(func, retries)



from guillotina import task_vars
from guillotina.tests.utils import get_container
from guillotina_amqp.tests.utils import _test_func
from guillotina_amqp.utils import add_task


async def test_list_tasks_returns_all_tasks(container_requester, dummy_request):
    async with container_requester as requester:
        task_vars.request.set(dummy_request)
        task_vars.db.set(requester.db)
        await get_container(requester=requester)

        # Add some tasks first
        t1 = await add_task(_test_func, 1, 2)
        t2 = await add_task(_test_func, 3, 4)

        resp, status = await requester("GET", "/db/guillotina/@amqp-tasks")
        assert status == 200
        assert len(resp) == 2
        assert t1.task_id in resp
        assert t2.task_id in resp


async def test_info_task(container_requester, dummy_request):
    async with container_requester as requester:
        task_vars.request.set(dummy_request)
        task_vars.db.set(requester.db)
        await get_container(requester=requester)

        # Add some tasks first
        t1 = await add_task(_test_func, 1, 2)

        # Check returns correctly if existing task_id
        resp, status = await requester(
            "GET", f"/db/guillotina/@amqp-tasks/{t1.task_id}"
        )
        assert status == 200
        assert resp["status"] == "scheduled"

        # Check returns 404 if task_id is not known
        resp, status = await requester("GET", "/db/guillotina/@amqp-tasks/foo")
        assert status == 404
        assert resp["reason"] == "Task not found"

        resp, status = await requester(
            "GET", "/db/guillotina/@amqp-tasks/task:db-guillotina-foobar"
        )
        assert status == 404
        assert resp["reason"] == "Task not found"


async def test_cancel_task(container_requester, dummy_request):

    async with container_requester as requester:
        task_vars.request.set(dummy_request)
        task_vars.db.set(requester.db)
        await get_container(requester=requester)

        # Add some tasks first
        t1 = await add_task(_test_func, 1, 2)

        # Check returns correctly if existing task_id
        resp, status = await requester(
            "DELETE", f"/db/guillotina/@amqp-tasks/{t1.task_id}"
        )
        assert status == 200
        assert resp is True

        # Check returns correctly if already canceled task_id
        resp, status = await requester(
            "DELETE", f"/db/guillotina/@amqp-tasks/{t1.task_id}"
        )
        assert status == 200
        assert resp is True

        # Check returns 404 is unknown task
        resp, status = await requester("DELETE", "/db/guillotina/@amqp-tasks/foo")
        assert status == 404

    task_vars.request.set(None)



from guillotina_amqp.decorators import task
from guillotina_amqp.interfaces import MessageType

import asyncio


async def _test_func(one, two, one_keyword=None):
    return one + two


async def _test_asyncgen(one, two, one_keyword=None):
    yield (MessageType.DEBUG, "Starting task")
    yield (MessageType.DEBUG, "Yellow")
    yield (MessageType.RESULT, one + two)


async def _test_asyncgen_invalid():
    yield (MessageType.DEBUG, 1, 2, 3)
    yield (MessageType.RESULT)


async def _test_asyncgen_doubley(arg1, arg2, arg3):
    yield (MessageType.RESULT, arg1)
    yield (MessageType.RESULT, arg2)
    yield (MessageType.DEBUG, "OK")
    yield (MessageType.RESULT, arg3)


@task
async def _test_long_func(duration):
    print("Started task")
    await asyncio.sleep(duration)
    print("Finished task")
    return "done!"


@task
async def _test_failing_func():
    raise Exception("Foobar")


@task
async def _decorator_test_func(one, two):
    return one + two



from aiohttp import test_utils
from aiohttp.helpers import noop
from datetime import datetime
from guillotina import glogging
from guillotina import task_vars as g_task_vars
from guillotina.auth.users import GuillotinaUser
from guillotina.auth.utils import set_authenticated_user
from guillotina.component import get_utility
from guillotina.interfaces import ACTIVE_LAYERS_KEY
from guillotina.interfaces import Allow
from guillotina.interfaces import IAnnotations
from guillotina.interfaces import IApplication
from guillotina.registry import REGISTRY_DATA_KEY
from guillotina.transactions import abort
from guillotina.transactions import commit
from guillotina.utils import get_dotted_name
from guillotina.utils import import_class
from guillotina.utils import resolve_dotted_name
from guillotina_amqp import task_vars
from guillotina_amqp.exceptions import ObjectNotFoundException
from guillotina_amqp.interfaces import ITaskDefinition
from guillotina_amqp.interfaces import MessageType
from guillotina_amqp.state import get_state_manager
from guillotina_amqp.state import update_task_running
from guillotina_amqp.utils import _run_object_task
from guillotina_amqp.utils import _yield_object_task
from multidict import CIMultiDict
from unittest import mock
from urllib.parse import urlparse
from zope.interface import alsoProvides

import inspect
import time
import yarl


logger = glogging.getLogger("guillotina_amqp.job")


def login_user(request, user_data):
    """Logs user in to guillotina so the job has the correct access

    """
    if "id" in user_data:
        user = GuillotinaUser(
            user_id=user_data["id"],
            groups=user_data.get("groups", []),
            roles={name: Allow for name in user_data["roles"]},
        )
        user.data = user_data.get("data", {})
        set_authenticated_user(user)
    else:
        set_authenticated_user(None)

    if user_data.get("Authorization"):
        # leave in for b/w compat, remove at later date
        request.headers["Authorization"] = user_data["Authorization"]
    for name, value in (user_data.get("headers") or {}).items():
        request.headers[name] = value


class EmptyPayload:
    async def readany(self):
        return bytearray()

    def at_eof(self):
        return True


class Job:
    """Job objects are responsible for running the actual functions that
    were configured for. They ack/nack rabbitmq when job is finished, and publish

    """

    def __init__(self, base_request, data, channel, envelope):
        if base_request is None:
            from guillotina.tests.utils import make_mocked_request

            base_request = make_mocked_request("POST", "/db")
        self.base_request = base_request
        self.data = data
        self.channel = channel
        self.envelope = envelope

        self.task = None
        self._state_manager = None
        self._started = time.time()

    @property
    def state_manager(self):
        if self._state_manager is None:
            self._state_manager = get_state_manager()
        return self._state_manager

    async def create_request(self):
        req_data = self.data["req_data"]
        url = req_data["url"]
        parsed = urlparse(url)
        dct = {
            "method": req_data["method"],
            "url": yarl.URL(url),
            "path": parsed.path,
            "headers": CIMultiDict(req_data["headers"]),
            "raw_headers": tuple(
                (k.encode("utf-8"), v.encode("utf-8"))
                for k, v in req_data["headers"].items()
            ),
        }

        message = self.base_request._message._replace(**dct)

        payload_writer = mock.Mock()
        payload_writer.write_eof.side_effect = noop
        payload_writer.drain.side_effect = noop

        protocol = mock.Mock()
        protocol.transport = test_utils._create_transport(None)
        protocol.writer = payload_writer

        request = self.base_request.__class__(
            message,
            EmptyPayload(),
            protocol,
            payload_writer,
            self.task,
            self.task._loop,
            client_max_size=self.base_request._client_max_size,
            state=self.base_request._state.copy(),
        )
        g_task_vars.request.set(request)
        request.annotations = req_data.get("annotations", {})

        if self.data.get("db_id"):
            root = get_utility(IApplication, name="root")
            db = await root.async_get(self.data["db_id"])
            g_task_vars.db.set(db)
            # Add a transaction Manager to request
            tm = db.get_transaction_manager()
            g_task_vars.tm.set(tm)
            # Start a transaction
            txn = await tm.begin()
            # Get the root of the tree
            context = await tm.get_root(txn=txn)

            if self.data.get("container_id"):
                container = await context.async_get(self.data["container_id"])
                if container is None:
                    raise Exception(
                        f'Could not find container: {self.data["container_id"]}'
                    )
                g_task_vars.container.set(container)
                annotations_container = IAnnotations(container)
                container_settings = await annotations_container.async_get(
                    REGISTRY_DATA_KEY
                )
                layers = container_settings.get(ACTIVE_LAYERS_KEY, [])
                for layer in layers:
                    try:
                        alsoProvides(request, import_class(layer))
                    except ModuleNotFoundError:
                        pass
                g_task_vars.registry.set(container_settings)
        return request

    async def __call__(self):
        # Clone request for task
        request = await self.create_request()
        try:
            result = await self.__run(request)
            try:
                # Finish and return result
                await commit()
                request.execute_futures()
            except Exception:
                logger.warning("Error commiting job", exc_info=True)
                raise
            return result
        except ObjectNotFoundException:
            # Tried to run a function on an object that no longer
            # exist in the database
            return None
        except Exception:
            raise
        finally:
            try:
                await abort()
            except Exception:
                logger.warning("Error aborting job", exc_info=True)

    def get_function_to_run(self):
        func = resolve_dotted_name(self.data["func"])
        if ITaskDefinition.providedBy(func):
            func = func.func
        if hasattr(func, "__real_func__"):
            # from decorators
            func = func.__real_func__
        return func

    @property
    def function_name(self):
        """
        """
        func = self.get_function_to_run()
        dotted_name = get_dotted_name(func)
        if func in [_run_object_task, _yield_object_task]:
            # Remove guillotina_amqp.utils part
            dotted_name = dotted_name.lstrip("guillotina_amqp.utils")
            # Get actuall callable that is passed as the first parameter
            # of func
            _callable = self.data.get("args", [""])[0]
            dotted_name += f"/{_callable}"

        return dotted_name

    async def __run(self, request):
        result = None
        task_id = self.data["task_id"]
        dotted_name = self.data["func"]
        logger.info(f"Running task: {task_id}: {dotted_name}")

        # Update status
        await update_task_running(self.state_manager, task_id)

        req_data = self.data["req_data"]
        if "user" in req_data:
            login_user(request, req_data["user"])

        # Parse the function to run
        func = self.get_function_to_run()

        #
        # Run the task coroutine
        #
        # Your coroutine can be a regular coroutine or an asynchronous
        # generator. If you use an asynchronous generator, your generator
        # should yield tuples of the form:
        #
        # (type, content)
        #
        # There are 2 different event types:
        #
        # type = MessageType.RESULT: task value yield
        # type = MessageType.DEBUG: task message, will be logged to the state manager3
        #
        # All the values yielded by the task are returned as a list to the
        # caller
        #

        task_vars.amqp_job.set(self)
        # Function is an async generator
        if inspect.isasyncgenfunction(func):
            async for status in func(*self.data["args"], **self.data["kwargs"]):
                if not isinstance(status, tuple) or len(status) != 2:
                    logger.debug(f"Job: invalid generator event: {status}")
                    continue

                msg_type, content = status

                if msg_type == MessageType.DEBUG:
                    # DEBUG message: record it in the task's event log
                    date_now = datetime.now().isoformat(" ", "milliseconds")
                    logger.debug(
                        f"Job {task_id}: function data {self.data}, got msg {content}"
                    )

                    # Update the status with new log
                    state = await self.state_manager.get(task_id)
                    state.setdefault("eventlog", [])
                    state["eventlog"].append([date_now, content])
                    await self.state_manager.update(task_id, state)

                elif msg_type == MessageType.RESULT:
                    # RESULT value yielded: accumulate all the
                    # generator results in a list
                    logger.debug(
                        f"Job {task_id}: function data {self.data}, got result {content}"
                    )
                    if result is None:
                        result = [content]
                    elif isinstance(result, list):
                        result.append(content)
                else:
                    # Unknown message: log and continue
                    logger.debug(
                        f"Job {task_id}: invalid generator event code {msg_type}"
                    )
                    continue
        else:
            # Regular coroutine
            result = await func(*self.data["args"], **self.data["kwargs"])
        task_vars.amqp_job.set(None)

        return result



from guillotina import app_settings
from guillotina import glogging
from guillotina.utils import resolve_dotted_name

import aioamqp
import aioamqp.exceptions
import asyncio


logger = glogging.getLogger("guillotina_amqp")


async def remove_connection(name="default"):
    """
    Purpose here is to close out a bad connection.
    Next time get_connection is called, a new connection will be established
    """
    amqp_settings = app_settings["amqp"]
    if "connections" not in amqp_settings:
        amqp_settings["connections"] = {}
    connections = amqp_settings["connections"]
    if name not in connections:
        return

    connection = connections.pop(name)
    try:
        await connection["protocol"].close(no_wait=True)
    except Exception:
        pass  # could already be closed
    try:
        connection["transport"].close()
    except Exception:
        pass


async def handle_connection_closed(name, protocol):
    try:
        # this just waits for the connection to close
        try:
            await protocol.wait_closed()
        except GeneratorExit:
            return
        # we just remove so next time get_connection is retrieved, a new
        # connection is retrieved.
        logger.warning(
            "Disconnect detected with rabbitmq connection, forcing reconnect"
        )
        await remove_connection(name)
    except Exception:
        logger.error("Error waiting for connection to close", exc_info=True)


async def heartbeat():
    while True:
        await asyncio.sleep(20)
        try:
            for name, connection in app_settings["amqp"].get("connections", {}).items():
                await connection["protocol"].send_heartbeat()
        except Exception:
            logger.error("Error sending heartbeat", exc_info=True)


async def get_connection(name="default"):
    amqp_settings = app_settings["amqp"]
    if "connections" not in amqp_settings:
        amqp_settings["connections"] = {}
    connections = amqp_settings["connections"]
    if name in connections:
        connection = connections[name]
        return connection["channel"], connection["transport"], connection["protocol"]
    channel, transport, protocol = await connect()

    connections[name] = {
        "channel": channel,
        "protocol": protocol,
        "transport": transport,
    }
    asyncio.ensure_future(handle_connection_closed(name, protocol))
    if amqp_settings.get("heartbeat_task") is None:
        logger.info("Starting amqp heartbeat")
        amqp_settings.update({"heartbeat_task": asyncio.ensure_future(heartbeat())})
    return channel, transport, protocol


async def connect(**kwargs):
    amqp_settings = app_settings["amqp"]
    conn_factory = resolve_dotted_name(
        amqp_settings.get("connection_factory", aioamqp.connect)
    )
    transport, protocol = await conn_factory(
        amqp_settings["host"],
        amqp_settings["port"],
        amqp_settings["login"],
        amqp_settings["password"],
        amqp_settings["vhost"],
        heartbeat=amqp_settings["heartbeat"],
        **kwargs
    )
    channel = await protocol.channel()
    return channel, transport, protocol



from guillotina_amqp.exceptions import TaskAccessUnauthorized
from guillotina_amqp.exceptions import TaskAlreadyAcquired
from guillotina_amqp.state import get_state_manager

import asyncio
import pytest


async def clear_cache(sm):
    await sm._clean()


async def test_update_and_get_should_match(configured_state_manager, loop):
    state_manager = get_state_manager(loop)
    await state_manager.update("foo", {"state": "bar"})
    data = await state_manager.get("foo")
    assert data["state"] == "bar"
    await clear_cache(state_manager)


async def test_acquire_should_block_further_acquires(configured_state_manager, loop):
    state_manager = get_state_manager(loop)
    FOREVER = 300
    await state_manager.acquire("foo", ttl=FOREVER)
    with pytest.raises(TaskAlreadyAcquired):
        await state_manager.acquire("foo", ttl=FOREVER)
    # Release it and acquire again
    await state_manager.release("foo")
    await state_manager.acquire("foo", ttl=FOREVER)
    await clear_cache(state_manager)


async def test_acquire_should_unblock_after_ttl(configured_state_manager, loop):
    state_manager = get_state_manager(loop)
    await state_manager.acquire("foo", ttl=1)
    with pytest.raises(TaskAlreadyAcquired):
        await state_manager.acquire("foo", ttl=300)
    # Wait for the ttl and check
    await asyncio.sleep(1.1)
    await state_manager.acquire("foo", ttl=1)
    await clear_cache(state_manager)


async def test_list_should_yield_all_items(configured_state_manager, loop):
    state_manager = get_state_manager(loop)
    tasks = set({"t1", "t2", "t3", "t4", "t5"})
    for task_id in tasks:
        await state_manager.update(task_id, {"state": f"{task_id} is OK!"})
    assert set([tid async for tid in state_manager.list()]) == tasks
    await clear_cache(state_manager)


async def test_cancel_should_put_tasks_in_cancelation_list(
    configured_state_manager, loop
):
    state_manager = get_state_manager(loop)
    await state_manager.update("foo", {"status": "bar"})
    canceled_list = [tid async for tid in state_manager.cancelation_list()]
    assert "foo" not in canceled_list
    await state_manager.cancel("foo")
    canceled_list = [tid async for tid in state_manager.cancelation_list()]
    assert "foo" in canceled_list
    await clear_cache(state_manager)


async def test_clean_cancel_should_clean_from_canceled_list(
    configured_state_manager, loop
):
    state_manager = get_state_manager(loop)
    await state_manager.cancel("foo")
    canceled_list = [tid async for tid in state_manager.cancelation_list()]
    assert "foo" in canceled_list
    await state_manager.clean_canceled("foo")
    canceled_list = [tid async for tid in state_manager.cancelation_list()]
    assert "foo" not in canceled_list
    await clear_cache(state_manager)


async def test_is_canceled_should_return_true_only_on_canceled_tasks(
    configured_state_manager, loop
):
    state_manager = get_state_manager(loop)
    await state_manager.cancel("foo")
    assert await state_manager.is_canceled("foo")
    assert not await state_manager.is_canceled("bar")
    await clear_cache(state_manager)


async def test_refresh_should_raise_if_task_is_not_yours(
    configured_state_manager, loop
):
    state_manager = get_state_manager(loop)
    state_manager.worker_id = "me"
    await state_manager.acquire("footask", ttl=120)
    await state_manager.refresh_lock("footask", ttl=20)
    with pytest.raises(TaskAccessUnauthorized):
        state_manager.worker_id = "another_person"
        await state_manager.refresh_lock("footask", 120)
    state_manager.worker_id = "me"
    await state_manager.refresh_lock("footask", 20)
    await clear_cache(state_manager)


async def test_release_should_raise_if_task_is_not_yours(
    configured_state_manager, loop
):
    state_manager = get_state_manager(loop)
    state_manager.worker_id = "me"
    await state_manager.acquire("footask", ttl=120)
    with pytest.raises(TaskAccessUnauthorized):
        state_manager.worker_id = "another_person"
        await state_manager.release("footask")
    state_manager.worker_id = "me"
    await state_manager.release("footask")
    await clear_cache(state_manager)


async def test_task_state_disappears_after_ttl(redis_state_manager, loop):
    state_manager = get_state_manager(loop)
    await state_manager.update("foo", {"state": "bar"}, ttl=2)
    data = await state_manager.get("foo")
    assert data["state"] == "bar"
    await asyncio.sleep(2)
    data = await state_manager.get("foo")
    assert not data
    await clear_cache(state_manager)



import asyncio
import uuid


class MockChannel:
    def __init__(self):
        self.published = []
        self.acked = []
        self.nacked = []

    async def publish(self, *args, **kwargs):
        self.published.append({"args": args, "kwargs": kwargs})

    async def basic_client_ack(self, *args, **kwargs):
        self.acked.append({"args": args, "kwargs": kwargs})

    async def basic_client_nack(self, *args, **kwargs):
        self.nacked.append({"args": args, "kwargs": kwargs})


class MockEnvelope:
    def __init__(self, uid):
        self.delivery_tag = uid


class MockAMQPChannel:
    def __init__(self, protocol):
        self.protocol = protocol
        self.consumers = []
        self.closed = False
        self.unacked_messages = []

    async def basic_qos(self, *args, **kwargs):
        pass

    async def exchange_declare(self, *args, **kwargs):
        pass

    async def queue_declare(self, queue_name, *args, **kwargs):
        if queue_name not in self.protocol.queues:
            self.protocol.queues[queue_name] = []
        if "arguments" in kwargs:
            arguments = kwargs["arguments"]
            if "x-dead-letter-routing-key" in arguments:
                self.protocol.dead_mapping[queue_name] = arguments[
                    "x-dead-letter-routing-key"
                ]

    async def queue_bind(self, *args, **kwargs):
        pass

    async def _basic_consume(self, handler, queue_name):
        while not self.closed:
            await asyncio.sleep(0.02)
            if queue_name not in self.protocol.queues:
                continue
            else:
                messages = self.protocol.queues[queue_name]
                self.protocol.queues[queue_name] = []
                self.unacked_messages.extend(messages)
                for message in messages:
                    await handler(
                        self,
                        message["message"],
                        MockEnvelope(message["id"]),
                        message["properties"],
                    )

    async def basic_client_ack(self, delivery_tag):
        for message in self.unacked_messages[:]:
            if delivery_tag == message["id"]:
                self.unacked_messages.remove(message)
                return message

    async def basic_client_nack(self, delivery_tag, multiple=False, requeue=False):
        message = await self.basic_client_ack(delivery_tag)
        if message:
            if requeue:
                # put back on same queue
                self.protocol.queues[message["queue"]].append(message)
            else:
                new_queue = self.protocol.dead_mapping[message["queue"]]
                self.protocol.queues[new_queue].append(message)

    async def basic_consume(self, handler, queue_name):
        self.consumers.append(
            asyncio.ensure_future(self._basic_consume(handler, queue_name))
        )

    async def publish(
        self, message, exchange_name=None, routing_key=None, properties={}
    ):
        if routing_key not in self.protocol.queues:
            self.protocol.queues[routing_key] = []
        self.protocol.queues[routing_key].append(
            {
                "id": str(uuid.uuid4()),
                "message": message,
                "properties": properties,
                "queue": routing_key,
            }
        )

    async def close(self):
        self.closed = True
        await asyncio.sleep(0.06)


class MockAMQPTransport:
    def __init__(self):
        pass

    def close(self):
        pass


class MockAMQPProtocol:
    def __init__(self):
        self.queues = {}
        self.dead_mapping = {}
        self.closed = False
        self.channels = []

    async def channel(self):
        channel = MockAMQPChannel(self)
        self.channels.append(channel)
        return channel

    async def wait_closed(self):
        while not self.closed:
            await asyncio.sleep(0.05)
        raise GeneratorExit()

    async def close(self):
        self.closed = True
        for channel in self.channels:
            await channel.close()

    async def send_heartbeat(self):
        pass


async def amqp_connection_factory(*args, **kwargs):
    return MockAMQPTransport(), MockAMQPProtocol()



from guillotina import app_settings
from guillotina import testing
from guillotina_amqp import amqp
from guillotina_amqp.worker import Worker
from pytest_docker_fixtures.containers.rabbitmq import rabbitmq_image

import os
import pytest
import uuid


base_amqp_settings = {
    "connection_factory": "guillotina_amqp.tests.mocks.amqp_connection_factory",
    "host": "localhost",
    "port": 5673,
    "login": "guest",
    "password": "guest",
    "vhost": "/",
    "heartbeat": 120,  # 2 minutes
    "exchange": "",
    "queue": "guillotina",
    "persistent_manager": "memory",
    "max_running_tasks": 10,
    "delayed_ttl_ms": 1000 * 60,  # <-- 1 minute
    "errored_ttl_ms": 1000 * 60 * 60 * 24 * 7,  # <-- 1 week
    "state_ttl": 10,
}

test_logger_settings = {
    "version": 1,
    "formatters": {
        "default": {
            "format": "%(asctime)s %(levelname)-8s %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "level": "DEBUG",
        }
    },
    "loggers": {
        "guillotina_amqp.state": {"level": "DEBUG", "handlers": ["console"]},
        "guillotina_amqp.job": {"level": "DEBUG", "handlers": ["console"]},
        "guillotina_amqp.amqp": {"level": "DEBUG", "handlers": ["console"]},
    },
}


def base_settings_configurator(settings):
    if "applications" in settings:
        settings["applications"].extend(
            [
                "guillotina_amqp",
                "guillotina_amqp.tests.package",
                "guillotina.contrib.redis",
            ]
        )
    else:
        settings["applications"] = [
            "guillotina_amqp",
            "guillotina_amqp.tests.package",
            "guillotina.contrib.redis",
        ]
    settings["amqp"] = base_amqp_settings
    settings["logging"] = test_logger_settings


testing.configure_with(base_settings_configurator)


@pytest.fixture("function")
def amqp_worker(loop):
    # Create worker
    _worker = Worker(loop=loop, check_activity=False)
    _worker.update_status_interval = 2
    loop.run_until_complete(_worker.start())

    yield _worker

    # Tear down worker
    for conn in [v for v in app_settings["amqp"].get("connections", []).values()]:
        loop.run_until_complete(conn["protocol"].close())
    _worker.cancel()
    app_settings["amqp"]["connections"] = {}


@pytest.fixture("function", params=[{"redis_up": True}, {"redis_up": False}])
def configured_state_manager(request, redis, dummy_request, loop):
    if request.param.get("redis_up"):
        # Redis
        app_settings["amqp"]["persistent_manager"] = "redis"
        app_settings["redis_prefix_key"] = f"amqpjobs-{uuid.uuid4()}-"
        app_settings.update(
            {
                "redis": {
                    "host": redis[0],
                    "port": redis[1],
                    "pool": {"minsize": 1, "maxsize": 5},
                }
            }
        )
        print("Running with redis")
        yield redis
    else:
        # Memory
        app_settings["amqp"]["persistent_manager"] = "memory"
        print("Running with memory")
        yield


@pytest.fixture("function")
def redis_state_manager(redis, dummy_request, loop):
    # Redis
    app_settings["amqp"]["persistent_manager"] = "redis"
    app_settings["redis_prefix_key"] = f"amqpjobs-{uuid.uuid4()}-"
    app_settings.update(
        {
            "redis": {
                "host": redis[0],
                "port": redis[1],
                "pool": {"minsize": 1, "maxsize": 5},
            }
        }
    )
    print("Running with redis")
    yield redis


@pytest.fixture("function")
async def amqp_channel():
    channel, transport, protocol = await amqp.get_connection()
    return channel


IS_TRAVIS = "TRAVIS" in os.environ


@pytest.fixture(scope="session")
def rabbitmq_runner():
    if IS_TRAVIS:
        host = "127.0.0.1"
        port = 5672
    else:
        host, port = rabbitmq_image.run()
    yield host, port
    if not IS_TRAVIS:
        rabbitmq_image.stop()


@pytest.fixture("function")
def rabbitmq_container(rabbitmq):
    app_settings["amqp"].update(
        {
            "connection_factory": "aioamqp.connect",
            "host": rabbitmq[0],
            "port": rabbitmq[1],
            "login": "guest",
            "password": "guest",
            "vhost": "/",
            "heartbeat": 120,  # 2 minutes
            "exchange": "guillotina",
            "queue": "guillotina",
        }
    )



from guillotina_amqp.utils import metric_measure


class MockPrometheusMetric:
    def __init__(self, labels=None):
        self.labels_called = False
        self.observe_called = False
        self._values = []
        self._labels = {l: None for l in labels or []}

    def labels(self, **labels):
        self.labels_called = True
        for l, v in labels.items():
            if l not in self._labels:
                raise Exception()
            self._labels[l] = v
        return self

    def observe(self, value):
        self.observe_called = True
        self._values.append(value)


def test_metric_measure():
    # Measure with None metric just returns
    metric_measure(None, "foo", "bar")

    # Measure fills labels and observe
    histogram = MockPrometheusMetric(["label1", "label2"])
    metric_measure(histogram, 20, {"label1": "foo", "label2": "bar"})
    assert histogram.labels_called
    assert histogram.observe_called
    assert histogram._labels["label1"] == "foo"
    assert histogram._labels["label2"] == "bar"
    assert histogram._values == [20]



from setuptools import find_packages
from setuptools import setup


try:
    README = open('README.rst').read()
except IOError:
    README = None

setup(
    name='guillotina_amqp',
    version='5.0.7',
    description='Integrate amqp into guillotina',
    long_description=README,
    install_requires=[
        'guillotina>=5.0.0.dev0,<6',
        'aioamqp',
        'lru-dict',
        'aioredis'
    ],
    author='Nathan Van Gheem',
    author_email='vangheem@gmail.com',
    url='https://github.com/guillotinaweb/guillotina_amqp',
    packages=find_packages(exclude=['demo']),
    include_package_data=True,
    tests_require=[
        'pytest'
    ],
    extras_require={
        'test': [
            'pytest>=4.1.1',
            'docker',
            'backoff',
            'psycopg2-binary',
            'pytest-asyncio>=0.8.0',
            'pytest-cov>=2.6.1',
            'coverage>=4.4',
            'pytest-docker-fixtures[rabbitmq]==1.3.5',
        ]
    },
    license='BSD',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
    ],
    entry_points={
    }
)



from .exceptions import TaskNotFoundException
from .state import get_state_manager
from .state import TaskState
from .utils import get_task_id_prefix
from guillotina import configure
from guillotina.interfaces import IContainer
from guillotina.response import HTTPNotFound


@configure.service(
    method="GET",
    name="@amqp-tasks",
    context=IContainer,
    permission="guillotina.ManageAMQP",
    summary="Returns the list of running tasks",
)
@configure.service(
    method="GET",
    name="@amqp-info",
    context=IContainer,
    permission="guillotina.ManageAMQP",
    summary="Deprecated: Returns the list of running tasks",
)
async def list_tasks(context, request):
    mngr = get_state_manager()
    ret = []
    task_prefix = get_task_id_prefix()
    async for el in mngr.list():
        if el.startswith(task_prefix):
            ret.append(el)
    return ret


@configure.service(
    method="GET",
    name="@amqp-tasks/{task_id}",
    context=IContainer,
    permission="guillotina.ManageAMQP",
    summary="Shows the info of a given task id",
)
@configure.service(
    method="GET",
    name="@amqp-info/{task_id}",
    context=IContainer,
    permission="guillotina.ManageAMQP",
    summary="Deprecated: Shows the info of a given task id",
)
async def info_task(context, request):
    task_prefix = get_task_id_prefix()
    if not request.matchdict["task_id"].startswith(task_prefix):
        return HTTPNotFound(content={"reason": "Task not found"})
    try:
        task = TaskState(request.matchdict["task_id"])
        return await task.get_state()
    except TaskNotFoundException:
        return HTTPNotFound(content={"reason": "Task not found"})


@configure.service(
    method="DELETE",
    name="@amqp-tasks/{task_id}",
    context=IContainer,
    permission="guillotina.ManageAMQP",
    summary="Cancel a specific task by id",
)
@configure.service(
    method="DELETE",
    name="@amqp-cancel/{task_id}",
    context=IContainer,
    permission="guillotina.ManageAMQP",
    summary="Deprecated: Cancel a specific task by id",
)
async def cancel_task(context, request):
    task_prefix = get_task_id_prefix()
    if not request.matchdict["task_id"].startswith(task_prefix):
        return HTTPNotFound(content={"reason": "Task not found"})
    task = TaskState(request.matchdict["task_id"])
    try:
        return await task.cancel()
    except TaskNotFoundException:
        return HTTPNotFound(content={"reason": "Task not found"})



from .decorators import object_task  # noqa
from .decorators import task  # noqa
from .utils import add_object_task  # noqa
from .utils import add_task  # noqa
from guillotina import configure


app_settings = {
    "amqp": {
        "connection_factory": "aioamqp.connect",
        "host": "localhost",
        "port": 5672,
        "login": "guest",
        "password": "guest",
        "vhost": "/",
        "heartbeat": 60 * 2,  # 2 minutes
        "exchange": "guillotina",
        "queue": "guillotina",
        "persistent_manager": "memory",
        "max_running_tasks": 200,
        "state_ttl": 60 * 60 * 24,  # 1 day
    },
    "commands": {"amqp-worker": "guillotina_amqp.commands.worker.WorkerCommand"},
}


def includeme(root):
    """
    custom application initialization here
    """
    configure.scan("guillotina_amqp.permissions")
    configure.scan("guillotina_amqp.api")



from guillotina import app_settings
from guillotina import configure
from guillotina import glogging
from guillotina.component import get_utility
from guillotina_amqp.exceptions import TaskAccessUnauthorized
from guillotina_amqp.exceptions import TaskAlreadyAcquired
from guillotina_amqp.exceptions import TaskNotFinishedException
from guillotina_amqp.exceptions import TaskNotFoundException
from guillotina_amqp.interfaces import IStateManagerUtility
from lru import LRU

import asyncio
import copy
import json
import time
import uuid


try:
    import aioredis
    from guillotina.contrib import redis
except ImportError:
    aioredis = None


logger = glogging.getLogger("guillotina_amqp.state")

DEFAULT_LOCK_TTL_S = 60 * 1  # 1 minute


class TaskStatus:
    SCHEDULED = "scheduled"
    CANCELED = "canceled"
    RUNNING = "running"
    FINISHED = "finished"
    ERRORED = "errored"


@configure.utility(provides=IStateManagerUtility, name="memory")
class MemoryStateManager:
    """
    Meaningless for anyting other than tests
    """

    def __init__(self, size=10):
        self.size = size
        self._data = LRU(self.size)
        self._locks = {}
        self._canceled = set()
        self.worker_id = uuid.uuid4().hex

    def set_loop(self, loop=None):
        pass

    async def update(self, task_id, data, ttl=None):
        # Updates existing data with new data
        existing = await self.get(task_id)
        existing.update(data)
        self._data[task_id] = existing

    async def get(self, task_id):
        return self._data.get(task_id, {})

    async def exists(self, task_id):
        return task_id in self._data

    async def list(self):
        for task_id in self._data.keys():
            yield task_id

    async def acquire(self, task_id, ttl):
        already_locked = await self.is_locked(task_id)
        if already_locked:
            raise TaskAlreadyAcquired(task_id)

        # Set new lock
        from guillotina_amqp.utils import TimeoutLock

        lock = TimeoutLock(self.worker_id)
        await lock.acquire(ttl=ttl)
        self._locks[task_id] = lock

    async def is_mine(self, task_id):
        if task_id not in self._locks:
            raise TaskNotFoundException(task_id)
        lock = self._locks[task_id]
        return lock.locked() and lock.worker_id == self.worker_id

    async def is_locked(self, task_id):
        if task_id not in self._locks:
            return False
        return self._locks[task_id].locked()

    async def release(self, task_id):
        if not await self.is_mine(task_id):
            # You can't refresh a lock that's not yours
            raise TaskAccessUnauthorized(task_id)
        # Release lock and pop it from data structure
        self._locks[task_id].release()
        self._locks.pop(task_id, None)

    async def refresh_lock(self, task_id, ttl):
        if task_id not in self._locks:
            raise TaskNotFoundException(task_id)

        if not await self.is_locked(task_id):
            raise Exception(f"Task {task_id} is not locked")

        if not await self.is_mine(task_id):
            # You can't refresh a lock that's not yours
            raise TaskAccessUnauthorized(task_id)

        # Refresh
        return await self._locks[task_id].refresh_lock(ttl)

    async def cancel(self, task_id):
        self._canceled.update({task_id})
        return True

    async def cancelation_list(self):
        canceled = copy.deepcopy(self._canceled)
        for task_id in canceled:
            yield task_id

    async def clean_canceled(self, task_id):
        try:
            self._canceled.remove(task_id)
            return True
        except KeyError:
            # Task id wasn't canceled
            return False

    async def is_canceled(self, task_id):
        return task_id in self._canceled

    async def _clean(self):
        self._data = LRU(self.size)
        self._locks = {}
        self._canceled = set()


_EMPTY = object()


def get_state_manager(loop=None):
    """Factory that gets the configured state manager.

    Currently we have two implementations: memory | redis
    """
    utility = get_utility(
        IStateManagerUtility, name=app_settings["amqp"]["persistent_manager"]
    )
    if loop:
        utility.set_loop(loop)
    return utility


@configure.utility(provides=IStateManagerUtility, name="redis")
class RedisStateManager:
    """Implementation of the IStateManagerUtility with Redis
    """

    def __init__(self, loop=None):
        self._cache_prefix = app_settings.get("redis_prefix_key", "amqpjobs-")
        self.loop = loop
        self._cache = None
        self.worker_id = uuid.uuid4().hex

    def lock_prefix(self, task_id):
        return f"{self._cache_prefix}lock:{task_id}"

    @property
    def cancel_prefix(self):
        return f"{self._cache_prefix}cancel"

    def set_loop(self, loop=None):
        if loop:
            self.loop = loop

    async def get_cache(self):
        if self._cache == _EMPTY:
            return None

        if aioredis is None:
            logger.warning("aioredis not installed")
            self._cache = _EMPTY
            return None

        if "redis" in app_settings:
            self._cache = aioredis.Redis((await redis.get_driver()).pool)
            return self._cache
        else:
            self._cache = _EMPTY
            return None

    async def update(self, task_id, data, ttl=None):
        """Updates the state of the task. ttl can be set to expire the state.
        """
        cache = await self.get_cache()
        if cache:
            value = data
            existing = await cache.get(self._cache_prefix + task_id)
            if existing:
                # Update existing with new data
                value = json.loads(existing)
                value.update(data)
            await cache.set(self._cache_prefix + task_id, json.dumps(value))
            if ttl:
                resp = await cache.expire(self._cache_prefix + task_id, ttl)
                return resp > 0

    async def get(self, task_id):
        cache = await self.get_cache()
        if cache:
            value = await cache.get(self._cache_prefix + task_id)
            if value:
                return json.loads(value)
        return {}

    async def exists(self, task_id):
        data = await self.get(task_id)
        return data is not None

    async def list(self):
        cache = await self.get_cache()
        async for key in cache.iscan(match=f"{self._cache_prefix}*"):
            yield key.decode().replace(self._cache_prefix, "")

    async def acquire(self, task_id, ttl):
        if await self.is_locked(task_id):
            raise TaskAlreadyAcquired(task_id)

        # Set the lock
        cache = await self.get_cache()
        resp = await cache.setnx(self.lock_prefix(task_id), self.worker_id)
        if not resp:
            raise Exception(f"Error acquiring {task_id}")

        # Need to set an expiration for the lock in redis at creation
        # time
        refreshed = await self.refresh_lock(task_id, ttl)
        return refreshed

    async def is_locked(self, task_id):
        cache = await self.get_cache()
        resp = await cache.get(self.lock_prefix(task_id))
        return resp is not None

    async def is_mine(self, task_id):
        cache = await self.get_cache()
        task_owner_id = await cache.get(self.lock_prefix(task_id))
        if not task_owner_id:
            return False
        return task_owner_id.decode() == self.worker_id

    async def release(self, task_id):
        if not await self.is_locked(task_id):
            # There is no lock, nothing to do
            return False
        if not await self.is_mine(task_id):
            # You can't release a task for which you don't own a lock
            raise TaskAccessUnauthorized
        cache = await self.get_cache()
        resp = await cache.delete(self.lock_prefix(task_id))
        return resp > 0

    async def refresh_lock(self, task_id, ttl):
        if not await self.is_locked(task_id):
            # There is no lock, nothing to do
            return False

        if not await self.is_mine(task_id):
            # You can't release a task for which you don't own a lock
            raise TaskAccessUnauthorized(task_id)

        cache = await self.get_cache()
        resp = await cache.expire(self.lock_prefix(task_id), ttl)
        return resp > 0

    async def cancel(self, task_id):
        cache = await self.get_cache()
        current_time = time.time()
        resp = await cache.zadd(self.cancel_prefix, current_time, task_id)
        return resp > 0

    async def cancelation_list(self):
        cache = await self.get_cache()
        async for val, score in cache.izscan(self.cancel_prefix):
            yield val.decode()

    async def clean_canceled(self, task_id):
        cache = await self.get_cache()
        resp = await cache.zrem(self.cancel_prefix, task_id)
        return resp > 0

    async def is_canceled(self, task_id):
        async for tid in self.cancelation_list():
            if tid == task_id:
                return True
        return False

    async def _clean(self):
        await self._cache.flushall()


class TaskState:
    """Wrapper around state_manager implementation so we can use it by
    just having a task_id
    """

    def __init__(self, task_id):
        self.task_id = task_id

    async def join(self, wait=0.5):
        util = get_state_manager()
        while True:
            data = await util.get(self.task_id)
            if not data:
                raise TaskNotFoundException(self.task_id)
            if data.get("status") in (
                TaskStatus.FINISHED,
                TaskStatus.ERRORED,
                TaskStatus.CANCELED,
            ):
                return data
            await asyncio.sleep(wait)

    async def get_state(self):
        util = get_state_manager()
        data = await util.get(self.task_id)
        if not data:
            raise TaskNotFoundException(self.task_id)
        return data

    async def get_status(self):
        """
        possible statuses:
        - scheduled
        - canceled
        - running
        - finished
        - errored
        """
        util = get_state_manager()
        data = await util.get(self.task_id)
        if not data:
            raise TaskNotFoundException(self.task_id)
        return data.get("status")

    async def get_result(self):
        util = get_state_manager()
        data = await util.get(self.task_id)
        if not data:
            raise TaskNotFoundException(self.task_id)
        if data.get("status") not in (TaskStatus.FINISHED, TaskStatus.ERRORED):
            raise TaskNotFinishedException(self.task_id)
        return data.get("result")

    async def cancel(self):
        util = get_state_manager()
        if await util.is_canceled(self.task_id):
            # Already canceled
            return True
        if not await util.exists(self.task_id):
            raise TaskNotFoundException
        # Cancel it
        return await util.cancel(self.task_id)

    async def acquire(self, ttl=DEFAULT_LOCK_TTL_S):
        util = get_state_manager()
        try:
            await util.acquire(self.task_id, ttl)
        except TaskAlreadyAcquired:
            logger.warning(f"Task {self.task_id} is already taken")
            return False
        else:
            return True

    async def refresh_lock(self, ttl=DEFAULT_LOCK_TTL_S):
        util = get_state_manager()
        await util.refresh_lock(self.task_id, ttl)

    async def release(self):
        util = get_state_manager()
        await util.release(self.task_id)

    async def is_canceled(self):
        util = get_state_manager()
        return await util.is_canceled(self.task_id)


async def update_task_status(
    state_manager, task_id, status, task=None, ttl=None, result=None, **kwargs
):
    if ttl is None:
        ttl = int(app_settings["amqp"]["state_ttl"])

    task_data = {"status": status}

    if result:
        task_data["result"] = result

    task_data.update(**kwargs)

    await state_manager.update(task_id, task_data, ttl=ttl)


async def update_task_errored(
    state_manager, task_id, task=None, ttl=None, result=None, **kwargs
):
    await update_task_status(
        state_manager,
        task_id,
        TaskStatus.ERRORED,
        task=task,
        ttl=ttl,
        result=result,
        error=task.print_stack() if task else None,
        **kwargs,
    )


async def update_task_finished(
    state_manager, task_id, task=None, ttl=None, result=None, **kwargs
):
    await update_task_status(
        state_manager,
        task_id,
        TaskStatus.FINISHED,
        task=task,
        ttl=ttl,
        result=result,
        **kwargs,
    )


async def update_task_scheduled(
    state_manager, task_id, task=None, ttl=None, result=None, **kwargs
):
    await update_task_status(
        state_manager,
        task_id,
        TaskStatus.SCHEDULED,
        task=task,
        ttl=ttl,
        result=result,
        **kwargs,
    )


async def update_task_canceled(
    state_manager, task_id, task=None, ttl=None, result=None, **kwargs
):
    await update_task_status(
        state_manager,
        task_id,
        TaskStatus.CANCELED,
        task=task,
        ttl=ttl,
        result=result,
        error=task.print_stack() if task else None,
        **kwargs,
    )


async def update_task_running(
    state_manager, task_id, task=None, ttl=None, result=None, **kwargs
):
    await update_task_status(
        state_manager,
        task_id,
        TaskStatus.RUNNING,
        task=task,
        ttl=ttl,
        result=result,
        **kwargs,
    )



from zope.interface import Attribute
from zope.interface import Interface


class MessageType(object):
    """Encapsulates constants message types"""

    RESULT = "result"
    DEBUG = "debug"


class IStateManagerUtility(Interface):
    async def update(task_id, data):
        """Updates data related to task id into state manager
        """
        raise NotImplementedError()

    async def get(self, task_id):
        """Gets whatever was stored in state manager for task_id
        """
        raise NotImplementedError()

    async def exists(self, task_id):
        """Returns whether a task id exists in the state manager
        """
        raise NotImplementedError()

    async def list(self):
        """
        Yields items from the list of stored items
        """
        raise NotImplementedError()

    async def is_mine(self, task_id):
        """Returns whether the worker has a lock on a given task_idq
        """
        raise NotImplementedError()

    async def cancel(self, task_id):
        """
        Sets task_id to the canceled set of tasks
        """
        raise NotImplementedError()

    async def acquire(self, task_id, ttl):
        """
        Get a lock on a certain task, by id.
        """
        raise NotImplementedError()

    async def release(self, task_id):
        """
        Release the lock for the specified task
        """
        raise NotImplementedError()

    async def refresh_lock(self, task_id, ttl):
        """
        Update lock TTL
        """
        raise NotImplementedError()

    async def cancelation_list(self):
        """
        Yields items from the canceled set
        """
        raise NotImplementedError()

    async def clean_canceled(self, task_id):
        """
        Removes a task from the canceled set
        """
        raise NotImplementedError()

    async def is_canceled(self, task_id):
        """
        Whether a task id has been cancelled
        """
        raise NotImplementedError()


class ITaskDefinition(Interface):
    func = Attribute("actual function to run")

    async def __call__(*args, _request=None, **kwargs):
        """
        schedule it
        """

    def after_request(*args, _request=None, _name=None, **kwargs):
        """
        schedule after request
        """

    def after_commit(*args, _request=None, **kwargs):
        """
        schedule after commit
        """
